/*
  AnthyYahooJIMService: Anthy-wrapper for YahooJIMService
  http://developer.yahoo.co.jp/jlp/JIMService/V1/conversion.html

  Original Copyright (C) 2008: Taku Kudo <taku@chasen.org>
  Modification Copyright (C) 2008: Toshiyuki Hanaoka <hanna@kmc.gr.jp>

  This program is free software: you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation, either version 3 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include <vector>
#include <sstream>
#include <iostream>
#include <fstream>
#include <iconv.h>
#include <netdb.h>
#include <unistd.h>
#include <errno.h>
#include <stdlib.h>
#include <anthy/anthy.h>
#include <anthy/dicutil.h>
#ifdef __CYGWIN__
#include <ctype.h>
#include <asm/byteorder.h>
#include <cygwin/in.h>
#endif // __CYGWIN__
#include <sys/types.h>
#include <sys/socket.h>

#ifndef YAHOO_APP_ID
#define YAHOO_APP_ID "YahooDemo"
#endif

namespace {

  // Configurable parameters
  static const char kAppID[]     = YAHOO_APP_ID;
  static const char kUserAgent[] = "Mozilla/4.0 (compatible; MSIE 6.0; Windows NT 5.1)";
  static const int  kTimeout     = 3;  // wait 3sec at most

  // modify them if need be
  static const char kHostName[]    = "jlp.yahooapis.jp";
  static const char kRequestBase[] ="/JIMService/V1/conversion?&"
    "response=katakana,hiragana,alphanumeric,half_katakana,half_alphanumeric&"
    "appid=";
  static const char kVersionString[] = "AnthyYahooJIMService";

  // debug file
  static std::ofstream dfs;

  // copied from mecab
  void iconv_wrapper(iconv_t ic,
                     const std::string &input,
                     std::string *output) {
    size_t ilen = input.size();
    size_t olen = ilen * 4;
    char *ibuf = const_cast<char *>(input.data());
    output->reserve(input.size() * 4);
    char *obuf_org = const_cast<char *>(output->data());
    char *obuf = obuf_org;
    std::fill(obuf, obuf + olen, 0);
    size_t olen_org = olen;
    ::iconv(ic, 0, &ilen, 0, &olen);  // reset iconv state
    while (ilen != 0) {
      if (::iconv(ic, (const char **)&ibuf, &ilen, &obuf, &olen)
          == (size_t)-1) {
        *output = input;
        return;
      }
    }
    output->assign(obuf_org, olen_org - olen);
    return;
  }

  // copied from mecab
  size_t utf8_mblen(const char *begin, const char *end) {
    const size_t len = static_cast<size_t>(end - begin);
    if (static_cast<unsigned char>(begin[0]) < 0x80) {
      return 1;
    } else if (len >= 2 && (begin[0] & 0xe0) == 0xc0) {
      return 2;
    } else if (len >= 3 && (begin[0] & 0xf0) == 0xe0) {
      return 3;
    } else if (len >= 4 && (begin[0] & 0xf8) == 0xf0) {
      return 4;
    } else if (len >= 5 && (begin[0] & 0xfc) == 0xf8) {
      return 5;
    } else if (len >= 6 && (begin[0] & 0xfe) == 0xfc) {
      return 6;
    }
    return 1;
  }

  void encodeURL(const std::string &input,
                 std::string *output) {
    static const char kDigits[] = "0123456789abcdef";
    const char *end = input.data() + input.size();
    output->clear();
    for (const char *begin = input.data(); begin < end; ++begin) {
#ifdef __CYGWIN__
      // in cygwin, isascii has not declaration for C++
      if(isascii(*begin) &&
#else
      if(::isascii(*begin) &&
#endif
          (::isdigit(*begin) || ::isalpha(*begin))) {
        *output += *begin;
      } else {
        *output += '%';
        *output += kDigits[(*begin >> 4) & 0x0f];
        *output += kDigits[*begin & 0x0f];
      }
    }
  }

  bool isReadTimeout(int socket, int timeout) {
    fd_set fds;
    struct timeval tv;
    FD_ZERO(&fds);
    FD_SET(socket, &fds);
    tv.tv_sec = timeout;
    tv.tv_usec = 0;
    if (::select(socket + 1,
                 &fds, 0, 0, &tv) < 0) {
      return true;
    }
    if (FD_ISSET(socket, &fds)) {
      return false;
    }
    return true;
  }

  // return $1 in "<foo>(.+)</foo>"
  bool findElement(const std::string &line,
                   const char *begin,
                   const char *end,
                   std::string *elm) {
    size_t b = line.find(begin);
    if (b == std::string::npos) {
      return false;
    }

    const size_t e = line.find(end);
    if (e == std::string::npos) {
      return false;
    }

    b += std::strlen(begin);
    if (b >= line.size() || b >= e) {
      return false;
    }

    *elm = line.substr(b, e - b);
    return true;
  }

  bool sendRequest(const std::string &raw_sentence,
                   std::string *xml,
                   bool is_predictive,
                   std::string *error) {
    std::string encoded;
    encodeURL(raw_sentence, &encoded);

    const char *appid = ::getenv("YAHOO_APP_ID");
    std::string id;
    if (appid) {
      id = std::string(appid);
    } else {
      id = std::string(kAppID);
    }

    std::string request = std::string(kRequestBase) +
      id + "&sentence=" + encoded;

    // predictive
    if (is_predictive) {
      request += "&mode=predictive";
    }

    // dictionary
    std::string dictionary = "&dictionary=default,name,place,zip,symbol";
    const char *dic = ::getenv("ANTHY_YAHOO_JIM_SERVICE_DIC");
    if (dic) {
      dictionary = "&dictionary=";
      dictionary += dic;
    }
    request += dictionary;

    // proxy
    std::string host = kHostName;
    std::string port = "80";
    const char *proxy = ::getenv("HTTP_PROXY");
    if (proxy) {
      host = proxy;
      const size_t pos = host.find(":");
      if (pos != std::string::npos) {
        port = host.substr(pos + 1, host.size() - pos);
        host = host.substr(0, pos);
      }
      // rewrite request
      request = std::string("http://") + kHostName + request;
    }

    // start connection
#ifndef __CYGWIN__
    struct addrinfo hints, *res;
    ::memset(&hints, 0, sizeof(hints));
    hints.ai_socktype = SOCK_STREAM;
    hints.ai_family = AF_INET;

    if (::getaddrinfo(host.c_str(), port.c_str(), &hints, &res) != 0) {
      *error = "::getaddrinfo() failed";
      return false;
    }

    const int sock = ::socket(res->ai_family,
                              res->ai_socktype,
                              res->ai_protocol);
#else // __CYGWIN__
    // cygwin doesn't have addrinfo
    const int sock = ::socket(AF_INET,
			      SOCK_STREAM,
			      0);

    std::istringstream istr(port);
    int port_num;
    istr>>port_num;

    struct sockaddr_in sa;
    sa.sin_family = AF_INET;
    sa.sin_port = htons(port_num);

    struct hostent *hoste;
    unsigned int **addrptr;
    hoste = gethostbyname(host.c_str());
    addrptr = (unsigned int **)hoste->h_addr_list;
    sa.sin_addr.s_addr = *(*addrptr);

#endif // __CYGWIN__
    if (sock < 0) {
      *error = "::socket() failed: ";
      *error += ::strerror(errno);
      return false;
    }

#ifdef __CYGWIN__
    if (::connect(sock, (struct sockaddr*) &sa, sizeof(sa)) < 0) {
#else // __CYGWIN__
    if (::connect(sock, res->ai_addr, res->ai_addrlen) < 0) {
#endif // __CYGWIN__
      *error = "::connect() failed: ";
      *error += ::strerror(errno);
      return false;
    }

    std::string message = "GET ";
    message += request;
    message += " HTTP/1.1\r\n";
    message += "UserAgent: ";
    message += kUserAgent;
    message += "\r\n";
    message += "Host: ";
    message += kHostName;
    message += "\r\n\r\n";

    const char *begin = message.data();
    const char *end = message.data() + message.size();
    while (begin < end) {
      const size_t p = ::send(sock,
                              begin,
                              static_cast<size_t>(end - begin),
                              0);
      begin += p;
    }

    static const size_t kBufferSize  = 8192;
    char buf[kBufferSize];
    while (true) {
      if (isReadTimeout(sock, kTimeout)) {
        *error = "connection timeout";
#ifndef __CYGWIN__
	::freeaddrinfo(res);
#endif // __CYGWIN__
        return false;
      }
      const size_t recv_size = ::recv(sock, buf, sizeof(buf), 0);
      xml->append(buf, recv_size);
      // fix it, it's too slow..
      if (xml->find("</ResultSet>") != std::string::npos) {
        break;
      }
    }
#ifndef __CYGWIN__
    ::freeaddrinfo(res);
#endif // __CYGWIN__

    // debug
    {
      // request contains AppID which must not be disclosed to users!
      // dfs << request << std::endl;
      dfs << *xml;
    }

    // check 200 response
    // proxy might return HTTP/1.0
    if ((xml->find("HTTP/1.1 200 OK") != 0) &&
	(xml->find("HTTP/1.0 200 OK") != 0)) { 
      *error = "Non-200 response";
      return false;
    }

    return true;
  }

  struct Segment {
    std::string text;
    std::string utf8_text;
    std::string alphanumeric;
    std::string halfalphanumeric;
    std::string katakana;
    std::string halfkatakana;
    std::string hiragana;
    std::vector<std::string> candidates;
    void clear() {
      text.clear();
      alphanumeric.clear();
      halfkatakana.clear();
      katakana.clear();
      halfkatakana.clear();
      hiragana.clear();
      candidates.clear();
    }
  };

  class Segments: public std::vector<Segment> {
  public:
    Segments(): ic_utf82euc_(0), ic_euc2utf8_(0) {
      ic_utf82euc_ = ::iconv_open("EUC-JP", "UTF-8");
      ic_euc2utf8_ = ::iconv_open("UTF-8",   "EUC-JP");
    }

    virtual ~Segments() {
      ::iconv_close(ic_utf82euc_);
      ::iconv_close(ic_euc2utf8_);
    }

    bool start(const std::string &raw_sentence,
               bool is_predictive = false) {
      clear();
      std::string xml, utf8, error;
      ::iconv_wrapper(ic_euc2utf8_, raw_sentence, &utf8);
      if (!sendRequest(utf8, &xml, is_predictive, &error)) {
        makeErrorResult(raw_sentence, error);
        return false;
      }

      if (!parseFromXML(xml)) {
        makeErrorResult(raw_sentence, "XML parse error");
        return false;
      }

      return true;
    }

    bool resizeSegments(size_t segment_index, int diff) {
      if (diff == 0) {
        return false;
      }

      if (segment_index >= size()) {
        return false;
      }

      if (segment_index == size() - 1 && diff > 0) {
        // do nothing
        return true;
      }

      std::vector<std::string> segs;

      if (diff > 0) {
        for (size_t i = 0; i <= segment_index; ++i) {
          segs.push_back((*this)[i].utf8_text);
        }

        std::string &last_text = segs.back();

        size_t idx = 0;
        std::string rem;
        for (idx = segment_index + 1; idx < size(); ++idx) {
          const char *begin = (*this)[idx].utf8_text.data();
          const char *end = begin + (*this)[idx].utf8_text.size();
          while (begin < end) {
            const size_t len = ::utf8_mblen(begin, end);
            last_text.append(begin, len);
            begin += len;
            --diff;
            if (diff == 0) {
              rem.assign(begin, static_cast<size_t>(end - begin));
              goto LAST;
            }
          }
        }

      LAST:
        if (!rem.empty()) {
          if (idx + 1 < size()) {
            rem += (*this)[idx + 1].utf8_text;
          }
          segs.push_back(rem);
          ++idx;
        }
        for (size_t i = idx + 1; i < size(); ++i) {
          segs.push_back((*this)[i].utf8_text);
        }
      } else {
        for (size_t i = 0; i < segment_index; ++i) {
          segs.push_back((*this)[i].utf8_text);
        }

        std::vector<const char *> start_pos;
        const char *begin = (*this)[segment_index].utf8_text.data();
        const char *end =  begin + (*this)[segment_index].utf8_text.size();
        while (begin < end) {
          start_pos.push_back(begin);
          const size_t len = ::utf8_mblen(begin, end);
          begin += len;
        }

        const int bpos = static_cast<int>(start_pos.size()) + diff;
        if (bpos > 0) {
          segs.push_back(std::string(start_pos[0],
                                  static_cast<size_t>(start_pos[bpos] -
                                                      start_pos[0])));
          segs.push_back(std::string(start_pos[bpos]));
        } else {
          segs.push_back((*this)[segment_index].utf8_text);
        }

        for (size_t i = segment_index + 1; i < size(); ++i) {
          segs.push_back((*this)[i].utf8_text);
        }
      }

      // make new rewquest and re-send it
      std::string request;
      for (size_t i = 0; i < segs.size(); ++i) {
        if (i != 0) {
          request += ' ';
        }
        request += segs[i];
      }

      return start(request, false);
    }

    void dump(std::string *output) {
      std::ostringstream os;
      for (size_t i = 0; i < this->size(); ++i) {
        os << "SegmentText: " << (*this)[i].text << std::endl;
        const std::vector<std::string> &candidates = (*this)[i].candidates;
        for (size_t j = 0; j < candidates.size(); ++j) {
          os << "\tCandidate[" << j << "]:"  << candidates[j] << std::endl;
        }
      }
      output->assign(os.str());
    }

  private:
    iconv_t ic_utf82euc_;
    iconv_t ic_euc2utf8_;

    void makeErrorResult(const std::string &raw_sentence,
                         const std::string &message) {
      clear();
      Segment seg;
      seg.text = raw_sentence;
      seg.candidates.push_back(message);
      push_back(seg);
    }

#define FIND_SPECIAL(begin, end, name) do {                             \
      if (!empty() && findElement(line, (begin), (end), &elm)) {        \
        Segment &seg = back();                                          \
        ::iconv_wrapper(ic_utf82euc_, elm, &seg.name);                 \
        continue;                                                       \
      } } while (0)

    bool parseFromXML(const std::string &xml) {
      std::istringstream is(xml);
      this->clear();

      std::string line, elm;
      while (std::getline(is, line)) {
        if (findElement(line, "<SegmentText>", "</SegmentText>", &elm)) {
          resize(size() + 1);
          Segment &seg = back();
          seg.clear();
          iconv_wrapper(ic_utf82euc_, elm, &seg.text);
          seg.utf8_text = elm;
          seg.candidates.clear();
          continue;
        }

        if (!empty() &&
            findElement(line, "<Candidate>", "</Candidate>", &elm)) {
          std::string cand;
          iconv_wrapper(ic_utf82euc_, elm, &cand);
          back().candidates.push_back(cand);
          continue;
        }

        FIND_SPECIAL("<Alphanumeric>", "</Alphanumeric>",
                     alphanumeric);
        FIND_SPECIAL("<HalfAlphanumeric>", "</HalfAlphanumeric>",
                     halfalphanumeric);
        FIND_SPECIAL("<Katakana>", "</Katakana>",
                     katakana);
        FIND_SPECIAL("<HalfKatakana>", "</HalfKatakana>",
                     halfkatakana);
        FIND_SPECIAL("<Hiragana>", "</Hiragana>",
                     hiragana);
      }

      return !empty();
    }
#undef FIND_SPECIAL
  };

  template <class T> class ObjectPool {
  public:
    T *get() {
      bool found = false;
      for (size_t i = 0; i < pool_.size(); ++i) {
        if (!pool_[i].second) {
          pool_[i].second = true;  // maked as used
          return pool_[i].first;
        }
      }

      if (!found) {
        pool_.push_back(std::make_pair<T *, bool>(new T, true));
        return pool_.back().first;
      }
      return 0;
    }

    void release(T *segment) {
      for (size_t i = 0; i < pool_.size(); ++i) {
        if (pool_[i].first == segment) {
          pool_[i].second = false;
          return;
        }
      }
    }

    void clear() {
      for (size_t i = 0; i < pool_.size(); ++i) {
        delete pool_[i].first;
      }
      pool_.clear();
    }

    // it's not thread safe
    static ObjectPool<T> *singleton() {
      static ObjectPool<T> mgr;
      return &mgr;
    }

  private:
    std::vector<std::pair<T *,  bool> > pool_;
    ObjectPool() {}
    virtual ~ObjectPool() { clear(); }
  };
}   // namespace

// it's just a placeholder
struct anthy_context {
  Segments segs;
};


//////////////////////////////////////////////////////////////////////////
int anthy_init(void) {
  const char *filename = ::getenv("ANTHY_YAHOO_JIM_SERVICE_LOGFILE");
  if (!filename) {
    filename = "/dev/null";
  }
  dfs.open(filename);
  return 0;
}

void anthy_quit(void) {
  dfs.close();
  ::ObjectPool<anthy_context>::singleton()->clear();
}

void anthy_conf_override(const char *, const char *) {}

int anthy_set_personality(const char *) {
  return 0;
}

anthy_context_t anthy_create_context(void) {
  return ::ObjectPool<anthy_context>::singleton()->get();
}

void anthy_reset_context(anthy_context_t ac) {
  if (!ac) {
    return;
  }
  ac->segs.clear();
}

void anthy_release_context(anthy_context_t ac) {
  ObjectPool<anthy_context>::singleton()->release(ac);
}

int anthy_set_string(anthy_context_t ac, const char *s) {
  if (!ac || !s) {
    return -1;
  }
  ac->segs.start(s, false);
  return 0;
}

void anthy_resize_segment(anthy_context_t ac, int nth, int resize) {
  if (!ac) {
    return;
  }
  ac->segs.resizeSegments(nth, resize);
  return;
}

int anthy_get_stat(anthy_context_t ac, struct anthy_conv_stat *s) {
  if (!ac) {
    return -1;
  }
  s->nr_segment = ac->segs.size();
  return 0;
}

int anthy_get_segment_stat(anthy_context_t ac, int nth,
                           struct anthy_segment_stat *s) {
  if (!ac || !s || nth < 0 || nth >= static_cast<int>(ac->segs.size())) {
    return -1;
  }
  s->nr_candidate = ac->segs[nth].candidates.size();
  s->seg_len = ac->segs[nth].text.size();
  return 0;
}

int anthy_get_segment(anthy_context_t ac, int nth_seg, int nth_cand,
                      char *buf, int buflen) {
  if (!ac || nth_seg < 0 || nth_seg >= static_cast<int>(ac->segs.size())) {
    return -1;
  }

  const Segment &seg = ac->segs[nth_seg];

  std::string result;
  if (nth_cand == NTH_UNCONVERTED_CANDIDATE) {
    result = seg.text;
#ifdef NTH_KATAKANA_CANDIDATE
  } else if (nth_cand == NTH_KATAKANA_CANDIDATE) {
    result = seg.katakana;
#endif
#ifdef NTH_HIRAGANA_CANDIDATE
  } else if (nth_cand == NTH_HIRAGANA_CANDIDATE) {
    result = seg.hiragana;
#endif
#ifdef NTH_HALFKANA_CANDIDATE
  } else if (nth_cand == NTH_HALFKANA_CANDIDATE) {
    result = seg.halfkatakana;
#endif
  } else if (nth_cand >= 0 &&
             nth_cand < static_cast<int>(seg.candidates.size())) {
    result = seg.candidates[nth_cand];
  } else {
    return -1;
  }

  // getting-length mode
  if (!buf) {
    return result.size();
  }

  if (static_cast<int>(result.size()) + 1 > buflen) {
    return -1;
  }

  ::memcpy(buf, result.data(), result.size());
  buf[result.size()] = '\0';

  return result.size();
}

int anthy_commit_segment(anthy_context_t ac, int s, int c) {
  if (!ac) {
    return -1;
  }
  return 0;
}

int anthy_set_prediction_string(anthy_context_t ac, const char* s) {
  if (!ac || !s) {
    return -1;
  }
  ac->segs.start(s, true);
  return 0;
}

int anthy_get_prediction_stat(anthy_context_t ac,
                              struct anthy_prediction_stat *ps) {
  if (!ac || !ps) {
    return -1;
  }
  if (ac->segs.size() != 1) {  // segments size should be 1
    return -1;
  }

  ps->nr_prediction = ac->segs[0].candidates.size();
  return 0;
}

int anthy_get_prediction(anthy_context_t ac, int nth, char *buf, int buflen) {
  if (!ac) {
    return -1;
  }
  if (ac->segs.size() != 1) {  // segments size should be 1
    return -1;
  }

  if (nth < 0 ||
      static_cast<int>(ac->segs[0].candidates.size()) <= nth) {
    return -1;
  }

  const std::string &result = ac->segs[0].candidates[nth];
  // getting-length mode
  if (!buf) {
    return result.size();
  }

  if (static_cast<int>(result.size()) + 1 > buflen) {
    return -1;
  }

  ::memcpy(buf, result.data(), result.size());
  buf[result.size()] = '\0';

  return result.size();
}

int anthy_commit_prediction(anthy_context_t ac , int nth) {
  return 0;
}

void anthy_print_context(anthy_context_t ac) {}

const char *anthy_get_version_string(void) {
  return kVersionString;
}

void anthy_set_logger(anthy_logger , int level) {}

int anthy_context_set_encoding(anthy_context_t ac, int encoding) {
  return 0;
}

int anthy_set_reconversion_mode(anthy_context_t ac, int mode) {
  return 0;
}

void anthy_dic_util_init(void) {}
void anthy_dic_util_quit(void) {}
void anthy_dic_util_set_personality(const char *) {}
const char *anthy_dic_util_get_anthydir(void) { return 0; }
int anthy_dic_util_set_encoding(int ) { return -1; }

void anthy_priv_dic_delete(void) {}
int anthy_priv_dic_select_first_entry(void) { return -1; }
int anthy_priv_dic_select_next_entry(void) { return -1; }
int anthy_priv_dic_select_entry(const char *) { return -1; }

char *anthy_priv_dic_get_index(char *buf, int len)  { return 0; }
int anthy_priv_dic_get_freq(void) { return 0; }
char *anthy_priv_dic_get_wtype(char *buf, int len) { return 0; }
char *anthy_priv_dic_get_word(char *buf, int len) { return 0; }

int anthy_priv_dic_add_entry(const char *yomi, const char *word,
                             const char *wt, int freq) { return -1; }

char *anthy_dic_search_words_file(const char *word) { return 0; }

// debugging
#if 0
int main() {
  anthy_context_t c = ::ObjectPool<anthy_context>::singleton()->get();
  ::Segments &segs = c->segs;
  std::string line;
  while (std::getline(std::cin, line)) {
    segs.start(line, false);
    std::string output;
    segs.dump(&output);
    std::cout << output;
    segs.resizeSegments(1, -3);
    segs.dump(&output);
    std::cout << output;
  }
  return 0;
}
#endif

